package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

//////////////////////////////////////////////////////////////////////////////
// TODO:
//
//  - fix public/package/private status
//

//////////////////////////////////////////////////////////////////////////////
// Optimizations to add
//
// ** NOTE: not all of these are appropriate for this class -- it is
//          simply a list of optimizations not implemented.  This
//          class is meant to remain simple and easy to understand;
//          optimizations which obscure that do not belong here (they
//          should go into the compiled version instead)

/** implements Tomita's Graph Structured Stack */
class GSS {

    public GSS() { }

    private Phase.Node[] reducing_list = null;

    /** corresponds to a positions <i>between tokens</i> the input stream; same as Tomita's U_i's */
    public class Phase implements Invokable<Parser.Table.State, Forest, GSS.Phase.Node> {

        /** the token immediately after this phase */
        public  final Token token;

        boolean reducing = false;

        /** currently this is necessary only for the code() hack -- it doesn't actually correspond to the input */
        private final int pos;

        /** FIXME */
        public  Forest.Ref finalResult = null;

        /** all nodes, keyed by the value returned by code() */
        private HashMap<Long,Phase.Node> hash    = new HashMap<Long,Phase.Node>();  /* ALLOC */

        /** the number of nodes in this phase */
        private int numNodes = 0;

        boolean closed = false;

        private Token.Location location;
        public Phase(Phase previous, Token token, Token.Location location) {
            this.pos = previous==null ? 0 : previous.pos+1;
            this.token = token;
            this.location = location;
        }

        public boolean isDone() { return token == null; }

        private String error = "generic syntax error";
        public void checkFailure() throws Parser.Failed {
            if (numNodes <= 0)
                throw new Parser.Failed(error, getLocation());
        }

        public Token.Location getLocation() { return location; }

        /** add a new node (merging with existing nodes if possible)
         *  @param parent             the parent of the new node
         *  @param result             the SPPF result corresponding to the new node
         *  @param state              the state that the new node is in
         *  @param fromEmptyReduction true iff this node is being created as a result of a reduction of length zero (see GRMLR paper)
         *  @param start              the earliest part of the input contributing to this node (used to make merging decisions)
         */
        public void newNode(Node parent, Forest pending, Parser.Table.State state, boolean fromEmptyReduction) {
            Node p = hash.get(code(state, parent==null?null:parent.phase()));
            if (p != null)  newNode2(p, parent, pending, state, fromEmptyReduction);
            else            newNode3(parent, pending, state, fromEmptyReduction);
        }
        private void newNode2(Node p, Node parent, Forest pending, Parser.Table.State state, boolean fromEmptyReduction) {
            p.holder.merge(pending);
            if (p.parents().contains(parent)) return;
            p.parents().add(parent, true);
            if (p!=parent && !fromEmptyReduction) p.queueReductions(parent);
        }
        private void newNode3(Node parent, Forest pending, Parser.Table.State state, boolean fromEmptyReduction) {
            do {
                if (token != null && state.canShift(token)) break;
                if (state.isAccepting()) break;
                if (token==null) break;
                if (!state.canReduce(token)) return;
                //if (count > 1) break;
                //if (r.numPop == 0) break;
                //r.reduce(pending, parent, null, Phase.this, null);
                //return;
            } while(false);

            Node n = new Node(parent, pending, state);  // ALLOC
            n.queueEmptyReductions();
            if (!fromEmptyReduction) n.queueReductions(parent);
        }

        
        /** perform all reduction operations */
        public void reduce() {
            reducing = true;
            if (reducing_list==null || reducing_list.length < hash.size())
                reducing_list = new Phase.Node[hash.size() * 4];
            Collection<Node> hv = hash.values();
            hv.toArray(reducing_list);
            int num = hv.size();
            for(int i=0; i<num; i++) {
                Node n = reducing_list[i];
                n.queueEmptyReductions();
                // INVARIANT: we never "see" a node until its parent-set is complete, modulo merges
            }
            for(int i=0; i<num; i++) {
                Node n = reducing_list[i];
                reducing_list[i] = null;
                n.queueReductions();
            }
        }

        public void invoke(Parser.Table.State st, Forest result, Node n) {
            next.newNode(n, result, st, true);
        }
        private Phase next = null;

        /** perform all shift operations, adding promoted nodes to <tt>next</tt> */
        public void shift(Phase next, Forest result) {
            this.next = next;
            closed = true;
            Forest res = null;
            boolean ok = false;
            for(Phase.Node n : hash.values()) {
                if (n.holder==null) continue;
                n.holder.resolve();
                if (token == null && n.state.isAccepting()) {
                    ok = true;
                    if (finalResult==null) finalResult = new Forest.Ref();
                    finalResult.merge(n.holder);
                }
                if (!n.holder.valid()) continue;
                if (token == null) continue;
                n.state.invokeShifts(token, this, result, n);
                /*
                for(Parser.Table.State st : n.state.getShifts(token)) {
                    if (res == null) res = result;
                    next.newNode(n, res, st, true, this);
                    ok = true;
                }
                */
            }

            if (!ok && token != null) {
                StringBuffer error = new StringBuffer();
                error.append("error: unable to shift token \"" + token + "\"\n");
                //error.append("  before: " +pendingReductions+ "\n");
                //error.append("  before: " +totalReductions+ "\n");
                //for(Phase.Node n : hash.values()) {
                //n.queueReductions();
                //n.queueEmptyReductions();
                //}
                //error.append("  after: " +pendingReductions+ "\n");
                //error.append("  candidate states:\n");
                //for(Phase.Node n : hash.values()) {
                    //for(Sequence.Position p : n.state) error.append("        " + p + "\n");
                    //error.append("        --\n");
                //for(Parser.Table.Reduction r : n.state.getReductions(token)) error.append("        " + r + "\n");
                    //error.append("        ==\n");
                //}
                next.error = error.toString();
            }

            // this massively improves GC performance
            hash = null;
        }

       
        // GSS Nodes //////////////////////////////////////////////////////////////////////////////

        /** a node in the GSS */
        public final class Node extends FastSet<Node> implements Invokable<Parser.Table.Reduction, Node, Node> {

            private Forest.Ref holder = null;
            private boolean allqueued = false;

            /** what state this node is in */
            public final Parser.Table.State state;

            /** which Phase this Node belongs to (node that Node is also a non-static inner class of Phase) */
            public Phase phase() { return Phase.this; }

            public  Forest.Ref holder() { return holder==null ? (holder = new Forest.Ref()) : holder; }
            public  Forest pending() { return Phase.this.closed ? holder().resolve() : holder; }
            public  FastSet<Node> parents() { return this; }

            public void queueReductions() {
                if (allqueued) return;
                allqueued = true;
                int where = parents().size();
                state.invokeReductions(token, this, this, null);
            }

            public void queueReductions(Node n2) {
                if (!allqueued) { queueReductions(); return; }
                state.invokeReductions(token, this, this, n2);
            }

            public final void invoke(Parser.Table.Reduction r, Node n, Node n2) {
                if (n==null) {
                    if (r.numPop==0) r.reduce(this);
                    return;
                }
                if (r.numPop==0) return;
                if (n2==null) r.reduce(n);
                else          r.reduce(n, n2);
            }
            public void queueEmptyReductions() {
                if (!reducing) return;
                state.invokeReductions(token, this, null, null);
            }

            private Node(Node parent, Forest pending, Parser.Table.State state) {
                this.state = state;
                Phase start = parent==null ? null : parent.phase();
                if (pending != null) this.holder().merge(pending);
                if (parent != null) parents().add(parent, true);
                if (Phase.this.hash.get(code(state, start)) != null) throw new Error("severe problem!");
                Phase.this.hash.put(code(state, start), this);
                Phase.this.numNodes++;
                if (parent==null) holder().valid = true; // hack to make sure that the "base" node is always considered valid
            }
        }

    }

    /** helper method */
    private static boolean equal(Object a, Object b) {
        if (a==null && b==null) return true;
        if (a==null || b==null) return false;
        return a.equals(b);
    }

    /** this is something of a hack right now */
    private static long code(Parser.Table.State state, Phase start) {
        return (((long)state.idx) << 32) | (start==null ? 0 : (start.pos+1));
    }
}
